// TODO: check that Modifiers.get( int ) argument doesn't contain SE modifiers

header
{
    package com.unitesk.javatesk;

    import java.io.File;
    import java.util.List;
    import java.util.EnumSet;
    import com.unitesk.java5.Message;
    import com.unitesk.atp.text.location.*;
    import com.unitesk.atp.messages.*;
    import com.unitesk.atpj5.tree.tool.antlr.*;
    import com.unitesk.java5.*;
    import com.unitesk.java5.Java5.*;
    import com.unitesk.javatesk.SeJava5.*;

    import static com.unitesk.java5.TokenUtils.getPosition;
}

class SeJava5Parser extends Java5Parser;
options
{
    k = 2;
}

tokens
{
    CASUAL="casual";
    COVERAGE="coverage";
    EXECUTE="execute";
    PRIME="prime";
    SPECIFICATION="specification";
    REFINES="refines";
    UP="up";
    DOWN="down";
    INVARIANT="invariant";
    PRE="pre";
    UPDATE="update";
    POST="post";
    MEDIATOR="mediator";
    IMPLEMENTATION="implementation";
    SCENARIO="scenario";
    STATE="state";
    ITERATE="iterate";
    SERIALIZE="serialize";
    CHOICE="choice";
    BRANCH="branch";
    MARK="mark";
    IRRELEVANT="irrelevant";
    TAUTOLOGY="tautology";
    THROWN="thrown";
    REACTION="reaction";
    FILTER="filter";
    EXPECT="expect";
    TIMEOUT="timeout";
    EVENT = "event";
}

{
    /**
      * Class containing boolean flag to be used in primary rule
      * for flagging situations: whether qualified identifier is
      * matched at primary expression start or not.
     */
    protected class IsPrimaryPrefixQID {
        public boolean flag = false;
    }

    /**
      * Method sets the flag to "true" if a primary expression
      * begins with a qualified identifier and return "true"
      * to allow its use as an expression inside of semantic
      * predicate.
      */
    protected boolean turnOnIsPrefixQID(IsPrimaryPrefixQID f) {
        f.flag = true;
        return true;
    }

    /**
      * Method sets the flag to "false" if a primary expression
      * doesn't begin with a qualified identifier and return "true"
      * to allow its use as an expression inside of semantic
      * predicate.
      */
    protected boolean turnOffIsPrefixQID(IsPrimaryPrefixQID f) {
        f.flag = false;
        return true;
    }

    /**
      * Counts the number of LTHN ("<") seen in the typeArguments production.
      * It is used in semantic predicates to ensure we have seen
      * enough closing '>' characters; which actually may have been
      * either GTHN (">"), RSHIFT (">>") or URSHIFT (">>>") tokens.
     */
    protected int ltCounter = 0;

    protected final Java5Factory nodeCreator = new Java5NodeFactory();
    protected final SeJava5Factory seNodeCreator = new SeJava5NodeFactory();
    protected final ListFactory arrayListFactory = new ArrayListFactory();

    protected final Java5Factory nullCreator = new Java5NullFactory();
    protected final SeJava5Factory seNullCreator = new SeJava5NullFactory();
    protected final ListFactory nullListFactory = new NullListFactory();

    boolean createTree = true;

    protected Java5Factory creator;
    protected SeJava5Factory seCreator;
    protected ListFactory listFactory;

    protected File file;
    protected MessageReceiverAdapter mbox;
    protected Java5StartPositionVisitor startVisitor;
    protected Java5EndPositionVisitor endVisitor;

    protected JavaDocFilter javadocFilter;

    public void setFile( File file )
    {
        this.file = file;
    }

    public void setMessageBox( MessageReceiverAdapter mbox )
    {
        this.mbox = mbox;
    }

    public void setStartVisitor( Java5StartPositionVisitor startVisitor )
    {
        this.startVisitor = startVisitor;
    }

    public void setEndVisitor( Java5EndPositionVisitor endVisitor )
    {
        this.endVisitor = endVisitor;
    }

    public void setJavaDocFilter( JavaDocFilter f )
    {
        this.javadocFilter = f;
    }
    
    protected void createTree( boolean create )
    {
        if( create )
        {
            creator =  nodeCreator;
            seCreator =  seNodeCreator;
            listFactory = arrayListFactory;
        } else {
            creator =  nullCreator;
            seCreator =  seNullCreator;
            listFactory = nullListFactory;
        }
        createTree = create;
    }

    protected void stopCreateTree()
    {
        createTree( false );
    }

    private String tokenText( Token t )
    {
        String text = t.getText();
        if( text != null )
        {
            return  "'" + text + "'";
        } else {
            eofFound = true;
            return "EOF";
        }
    }

    private boolean eofFound = false;

    public void reportError( RecognitionException ex )
    {
        stopCreateTree();
    
        if( eofFound ) return;

        if( ex instanceof NoViableAltException )
        {
            NoViableAltException e = (NoViableAltException)ex;
            String s = tokenText( e.token );
            if( !eofFound )
            {
                try
                {
                    s += ' ' + tokenText( LT(2) );
                }
                catch( TokenStreamException tse ) {}
            }
            mbox.error( file
                      , new Position( ex.getLine()
                                    , ex.getColumn()
                                    )
                      , new Message.UnexpectedToken( s )
                      );
        } else
        if(    ex instanceof MismatchedTokenException
            && ((MismatchedTokenException)ex).mismatchType == MismatchedTokenException.TOKEN
          )
        {
            MismatchedTokenException e = (MismatchedTokenException)ex;
            mbox.error( file
                      , new Position( ex.getLine()
                                    , ex.getColumn()
                                    )
                      , new Message.MismatchedToken( _tokenNames[e.expecting], tokenText( e.token ) )
                      );
        } else {
            mbox.error( file
                      , new Position( ex.getLine()
                                    , ex.getColumn()
                                    )
                      , new Message.SyntaxError( ex.getMessage() )
                      );
        }
    }


    private void removeInappropriates( List<Token> optModifierList, java.util.Map<Modifier, EnumSet<Modifier>> inappropriatesMethodModifiers )
    {
        if ( optModifierList == null )
            return;

        java.util.Map<Modifier, Token> savedModifiers = TokenUtils.getModifiersAsMap( optModifierList );
        java.util.Map<Modifier, EnumSet<Modifier>> errorModifierPairs =
            Modifier.removeInappropriates(
                    optModifierList,
                    inappropriatesMethodModifiers );
        java.util.Set<Token> forDelete = new java.util.HashSet<Token>();
        for( Modifier modifier : errorModifierPairs.keySet() )
        {
            Token mToken = savedModifiers.get( modifier );

            EnumSet<Modifier> errModifiers = errorModifierPairs.get( modifier );
            for( Modifier m : errModifiers )
            {
                mbox.error( file
                      , getPosition( mToken )
                      , new Message.IllegalModifiersCombination( modifier.toString(), m.toString() )
                      );
                forDelete.add( savedModifiers.get( m ) );
            }
        }
        optModifierList.removeAll( forDelete );
    }

    protected static final java.util.Map<Modifier, EnumSet<Modifier>>
        methodDeclarationInappropriatesModifiers = new java.util.HashMap<Modifier, EnumSet<Modifier>>();

    static {
            methodDeclarationInappropriatesModifiers.put(Modifier.ABSTRACT, EnumSet.of(Modifier.STATIC, Modifier.STRICTFP, Modifier.PRIVATE));
            methodDeclarationInappropriatesModifiers.put(Modifier.STATIC, EnumSet.of(Modifier.ABSTRACT));
            methodDeclarationInappropriatesModifiers.put(Modifier.STRICTFP, EnumSet.of(Modifier.ABSTRACT));
            methodDeclarationInappropriatesModifiers.put(Modifier.PRIVATE, EnumSet.of(Modifier.ABSTRACT));
    }
}

/**
 * JavaTESK modifier.
 */
seModifier returns [Token modifier = null]
:
    { modifier = LT(1); }
    ( PRIME )
;

/**
 * Extention point: Java modifier or JavaTESK modifier.
 */
modifier returns [Token modifier = null]
:
      modifier = javaModifier
    | modifier = seModifier
;

/**
 * Optional sequence of modifiers and annotations.
 * @param  optAnnotationList   list to add annotations to
 * @param  optModifierList     list to add modifiers to
 * @return                     the first token of this sequence or null if it is empty
 */
modifiers[List<Annotation> optAnnotationList, List<Token> optModifierList]
{
    Token modifier = null;
    Annotation annotation = null;
    java.util.Map<String, Token> nameModifierMap = null;
    boolean cTree = createTree;
    java.util.List<Modifier> visitedModifiers = null;
    java.util.List<SeModifiers> visitedSeModifiers = null;
    java.util.Set<Modifier> modifiersForDelete = null;
}
:
    {
    	nameModifierMap = new java.util.HashMap<String,Token>();
	    visitedModifiers = new java.util.ArrayList<Modifier>();
	    visitedSeModifiers = new java.util.ArrayList<SeModifiers>();
    	modifiersForDelete = new java.util.HashSet<Modifier>();
    	createTree( true );
    }
    ( // turn off warnings since the semantic check for
      // "@interface" solves the non-determinism
      options { generateAmbigWarnings=false; }:
        modifier = modifier
        {
            String m = modifier.getText();
            Modifier mod = TokenUtils.tokenToModifier( modifier );
            // mod == null means that it isn't java modifier (it's possible in language extentions)
            if( mod != null ) 
            {
	            if( nameModifierMap.containsKey( m ) )
	            {
		            // taking off duplicates (public public)
	                mbox.error( file
	                          , getPosition( modifier )
	                          , new Message.DuplicateModifier
	                                ( m, getPosition( nameModifierMap.get( m ) ) )
	                          );
	            } else if ( ! mod.corresponds( visitedModifiers ) )
	            {
		            // taking off inappropriate modifiers ( public private )
		            EnumSet<Modifier> currentInappropriates = EnumSet.copyOf( visitedModifiers );
		            currentInappropriates.retainAll( mod.getInappropriates() );
		            modifiersForDelete.addAll( currentInappropriates );
	                nameModifierMap.put( m, modifier );
	                visitedModifiers.add( mod );
	                mbox.error( file
	                          , getPosition( modifier )
	                          , new Message.IllegalModifiersCombination( m, currentInappropriates.iterator().next().toString() )
	                          );
	            } else {
	                nameModifierMap.put( m, modifier );
	                visitedModifiers.add( mod );
	                //listFactory.add( optModifierList, modifier );
			    	createTree( true );
	            }
	        }
	        else
			{
				SeModifiers semod = TokenUtils.tokenToSeModifier( modifier );
			    nameModifierMap.put( semod.toString(), modifier );
			    visitedSeModifiers.add( semod );
			    //listFactory.add( optModifierList, modifier );
			   	createTree( true );
			}
        }
      | // Semantic check that "@interface" is not been matching
        // as it is not an annotation but annotation type declaration
        { LA(1) == AT && LA(2) != INTERFACE }?
        annotation = annotation
        {
            listFactory.add( optAnnotationList, annotation );
            createTree( true );
        }
    )*
    {
    	for( Modifier m1 : modifiersForDelete )
    		visitedModifiers.remove( m1 );
    		
    	if ( ! visitedModifiers.isEmpty() )
    	{
    		for( Modifier m2 : visitedModifiers )
    		{
    			listFactory.add( optModifierList, nameModifierMap.get( m2.toString() ) );
    		}
    	}
    	
    	if ( ! visitedSeModifiers.isEmpty() )
		{
			for( SeModifiers m2 : visitedSeModifiers )
			{
				listFactory.add( optModifierList, nameModifierMap.get( m2.toString() ) );
			}
		}
    		
        createTree( cTree );
    }
;

/**
 * Extention point: Java type declaration or JavaTESK type declararion
 */
typeDeclaration[ Token startToken
               , List<Annotation> optAnnotationList
               , List<Token> optModifierList
               ]
    returns [TypeDeclaration typeDeclaration = null]
:
      typeDeclaration = javaTypeDeclaration[startToken,optAnnotationList,optModifierList]
    | // specification class
      typeDeclaration = seSpecificationClassDeclaration[startToken,optAnnotationList,optModifierList]
    | // mediator class
      typeDeclaration = seMediatorClassDeclaration[startToken,optAnnotationList,optModifierList]
    | // scenario class
      typeDeclaration = seScenarioClassDeclaration[startToken,optAnnotationList,optModifierList]
    | // reaction mediator class
      typeDeclaration = seReactionMediatorClassDeclaration[startToken,optAnnotationList,optModifierList]
;

/**
 * Specification class declaration.
 */
seSpecificationClassDeclaration[ Token startToken
                               , List<Annotation> optAnnotationList
                               , List<Token> optModifierList
                               ]
    returns [SE_SpecificationClassDeclaration typeDeclaration = null]
{
    Identifier name = null;
    TypeParameters optTypeParameters = null;
    QualifiedParameterizedIdentifier optSuperClass = null;
    List<QualifiedParameterizedIdentifier> optInterfaceList = null;
    SE_ClassRefinement classRefinement = null;
    List<SE_ClassRefinement> optClassRefinementList = null;
    TypeBody typeBody = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    SPECIFICATION CLASS
    name = identifier
    { cTree = createTree; }
    optTypeParameters = optTypeParameters
    ( EXTENDS
      optSuperClass = qualifiedParameterizedIdentifier[false]
    )?
    {
        createTree( true );
    }
    (  IMPLEMENTS optInterfaceList = optInterfaceList
     | { optInterfaceList = listFactory.create()/*Type*/; }
    )
    {
        createTree( true );
        optClassRefinementList = listFactory.create();
    }
    ( classRefinement = seClassRefinement
        {
            listFactory.add( optClassRefinementList, classRefinement );
            createTree( true );
        }
    )*
    {
        createTree( cTree );
    }
    // This is the body of a class.
    // It can contain class fields and extra semicolons.
    typeBodyStartToken:LBRACE
    typeBody = typeBody[typeBodyStartToken]
    {
        Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        typeDeclaration = seCreator.createSE_SpecificationClassDeclaration
                                    ( getPosition( startToken )
                                    , getPosition( endToken )
                                    , typeBody
                                    , javadoc
                                    , optAnnotationList
                                    , modifiers
                                    , name
                                    , optTypeParameters
                                    , optSuperClass
                                    , optInterfaceList
                                    , optClassRefinementList
                                    );
    }
    RBRACE
;

/**
 * 'refines' clause of specification class.
 */
seClassRefinement returns [SE_ClassRefinement seClassRefinement = null]
{
    QualifiedParameterizedIdentifier refinedClass;
    QualifiedParameterizedIdentifier upMediator;
    QualifiedParameterizedIdentifier downMediator;
}
:
    startToken:REFINES
    refinedClass = qualifiedParameterizedIdentifier[false]
    UP
    upMediator = qualifiedParameterizedIdentifier[false]
    DOWN
    downMediator = qualifiedParameterizedIdentifier[false]
    {
        seClassRefinement = seCreator.createSE_ClassRefinement( getPosition( startToken ), refinedClass, upMediator, downMediator );
    }
;

/**
 * Extention point to define extra categories of member declarations.
 */

memberDeclaration returns [MemberDeclaration member = null]
{
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    Token startToken = null;
}
:
    {
        optAnnotationList = listFactory.create()/*Annotation*/;
        optModifierList = listFactory.create()/*Token*/;
        startToken = LT(1);
    }
    modifiers[optAnnotationList,optModifierList]
    (
        member = seInvariantDeclaration[startToken,optAnnotationList,optModifierList]
      | { LA(3) != CLASS }?
        member = seReactionMediatorMethodDeclaration[startToken,optAnnotationList,optModifierList]
      | { LA(2) != CLASS }?
        member = seOperationSpecification[startToken,optAnnotationList,optModifierList]
      | member = seImplementationInitializer[startToken,optAnnotationList,optModifierList]
      | member = seImplementationFieldDeclaration[startToken,optAnnotationList,optModifierList]
      | { LA(2) != CLASS }?
        member = seMediatorMethodDeclaration[startToken,optAnnotationList,optModifierList]
      | member = seUpdateBlockDeclaration[startToken,optAnnotationList,optModifierList]
      | { LA(2) != CLASS }?
        member = seScenarioMethodDeclaration[startToken,optAnnotationList,optModifierList]
      | member = seScenarioStateBlock[startToken,optAnnotationList,optModifierList]
      | member = annotatedMemberDeclaration[startToken,optAnnotationList,optModifierList]
    )
;

/**
 * Invariant declaration.
 */
seInvariantDeclaration[ Token startToken
                      , List<Annotation> optAnnotationList
                      , List<Token> optModifierList
                      ]
    returns [SE_InvariantDeclaration invariantDeclaration = null]
{
    Identifier name = null;
    Block body = null;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    INVARIANT
    name = identifier
    LPARENT
    ( description:STRING_LITERAL )?
    RPARENT
    body = block
    {
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        EnumSet<SeModifiers> seModifiers = TokenUtils.getSeModifiers( optModifierList );

        invariantDeclaration
            = seCreator.createSE_InvariantDeclaration
                        ( getPosition( startToken )
                        , body != null ? body.getEndPosition() : null
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , seModifiers
                        , name
                        , description != null ? description.getText() : null
                        , body
                        );
    }
;

/**
 * Specification operation (method or reaction) declaration.
 */
seOperationSpecification[ Token startToken
                        , List<Annotation> optAnnotationList
                        , List<Token> optModifierList
                        ]
    returns [SE_OperationSpecification seOperationSpecification = null]
{
    boolean reaction = false;
    OptionalType type = null;
    TypeParameters optTypeParameters = null;
    MethodHeader methodHeader = null;
    SE_OperationRefinement seOperationRefinement = null;
    List<SE_OperationRefinement> optOperationRefinementList = null;
    BlockStatement statement = null;
    List<BlockStatement> optCommonStatementList = null;
    SE_PreCondition optPreCondition = null;
    SE_PostCondition optPostCondition = null;
    List<BlockStatement> optImplementationStatementList = null;
    boolean cTree = createTree;
    boolean c = false;
    String javadoc = null;
}
:
{
    createTree( true );
    optOperationRefinementList = listFactory.create();
    optCommonStatementList = listFactory.create();
    optImplementationStatementList = listFactory.create();
    createTree = cTree;
}
    ( REACTION { reaction = true; } )?
    { javadoc = javadocFilter.getComment( startToken ); }
    SPECIFICATION
    optTypeParameters = optTypeParameters
    type = optionalType[false]
    methodHeader = methodHeader[optTypeParameters,type]
    { c = createTree; createTree( true ); }
    (
      seOperationRefinement = seOperationRefinement
      {
        listFactory.add( optOperationRefinementList, seOperationRefinement );
        createTree( true );
      }
      ( COMMA seOperationRefinement = seOperationRefinement
        {
            listFactory.add( optOperationRefinementList, seOperationRefinement );
            createTree( true );
        }
      )*
    )?
    { cTree = c; createTree( c ); }
    LBRACE
    { c = createTree; createTree( true ); }
    ( options { greedy = true; }:
      { LA(1) != PRE }? statement = blockStatement
      {
        listFactory.add( optCommonStatementList, statement );
        createTree( true );
      }
    )*
    { cTree = c; createTree( c ); }
    ( ( PRE LBRACE ) => optPreCondition = sePreCondition )?
    ( optPostCondition = sePostCondition )?
    { c = createTree; createTree( true ); }
    ( statement = blockStatement
        {
            listFactory.add( optImplementationStatementList, statement );
            createTree( true );
        }
    )*
    { cTree = c; createTree( c ); }
    endToken:RBRACE
    {
        // TODO: check that optTypeParameters is null

        removeInappropriates( optModifierList, methodDeclarationInappropriatesModifiers );
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );

        if(    optPreCondition == null
            && optPostCondition == null
            && optImplementationStatementList != null
            && optCommonStatementList != null
          )
        {
            optImplementationStatementList.addAll( optCommonStatementList );
            optCommonStatementList.clear();
        }

        if( reaction )
        {
            seOperationSpecification
                = seCreator.createSE_ReactionSpecification
                            ( getPosition( startToken )
                            , getPosition( endToken )
                            , javadoc
                            , optAnnotationList
                            , modifiers
                            , methodHeader
                            , optOperationRefinementList
                            , optCommonStatementList
                            , optPreCondition
                            , optPostCondition
                            , optImplementationStatementList
                            );
        } else {
            seOperationSpecification
                = seCreator.createSE_MethodSpecification
                            ( getPosition( startToken )
                            , getPosition( endToken )
                            , javadoc
                            , optAnnotationList
                            , modifiers
                            , methodHeader
                            , optOperationRefinementList
                            , optCommonStatementList
                            , optPreCondition
                            , optPostCondition
                            , optImplementationStatementList
                            );
        }
    }
;

/**
 * 'refines' clause of specification method.
 */
seOperationRefinement returns [SE_OperationRefinement seOperationRefinement = null]
{
    QualifiedParameterizedIdentifier optRefinedClass = null;
    Block optUpBlock = null;
    Block optDownBlock = null;
    Block optUpdateBlock = null;
}
:
    startToken:REFINES
    (
        endToken:SUPER
        { seOperationRefinement = seCreator.createSE_SuperOperationRefinement( getPosition( startToken ), getPosition( endToken ) ); }
      | (
          (   optRefinedClass = qualifiedParameterizedIdentifier[false] ( UP optUpBlock = block )?
            | UP optUpBlock = block
          )
          ( DOWN optDownBlock = block )?
          ( UPDATE optUpdateBlock = block )?
        )
        {
            seOperationRefinement = seCreator.createSE_GeneralOperationRefinement
                        ( getPosition( startToken ), optRefinedClass, optUpBlock, optDownBlock, optUpdateBlock );
        }
    )
;

/**
 * Precondition of specification method.
 */
sePreCondition returns [SE_PreCondition preCondition = null]
{
    Block body = null;
}
:
    startToken:PRE body = block
    {
        preCondition = seCreator.createSE_PreCondition( getPosition( startToken ), body );
    }
;

/**
 * Postcondition of specification method.
 */
sePostCondition returns [SE_PostCondition postCondition = null]
{
    Block body = null;
}
:
    startToken:POST body = block
    {
        postCondition = seCreator.createSE_PostCondition( getPosition( startToken ), body );
    }
;

// --------------------------------------------------------------
//     m e d i a t o r   c l a s s
// --------------------------------------------------------------

/**
 * Mediator class declaration.
 */
seMediatorClassDeclaration[ Token startToken
                          , List<Annotation> optAnnotationList
                          , List<Token> optModifierList
                          ]
    returns [SE_MediatorClassDeclaration typeDeclaration = null]
{
    Identifier name = null;
    TypeParameters optTypeParameters = null;
    QualifiedParameterizedIdentifier optSuperClass = null;
    List<QualifiedParameterizedIdentifier> optInterfaceList = null;
    TypeBody typeBody = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); } 
    MEDIATOR CLASS
    name = identifier
    { cTree = createTree; }
    optTypeParameters = optTypeParameters
    { createTree( true ); }
    ( EXTENDS
      optSuperClass = qualifiedParameterizedIdentifier[false]
    )?
    {
        createTree( true );
    }
    (   IMPLEMENTS optInterfaceList = optInterfaceList
      |
        {
            optInterfaceList = listFactory.create()/*Type*/;
        }
    )
    {
        createTree( cTree );
    }
    typeBodyStartToken:LBRACE
    typeBody = typeBody[typeBodyStartToken]
    {
        Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        typeDeclaration = seCreator.createSE_MediatorClassDeclaration
                                    ( getPosition( startToken )
                                    , getPosition( endToken )
                                    , typeBody
                                    , javadoc
                                    , optAnnotationList
                                    , modifiers
                                    , name
                                    , optTypeParameters
                                    , optSuperClass
                                    , optInterfaceList
                                    );
    }
    RBRACE
;

/**
 * Implementation field declaration.
 */
seImplementationFieldDeclaration[ Token startToken
                                , List<Annotation> optAnnotationList
                                , List<Token> optModifierList
                                ]
    returns [SE_ImplementationFieldDeclaration seImplementationFieldDeclaration = null]
{
    Type type = null;
    List<VariableDeclarator> variableDeclaratorList = null;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    IMPLEMENTATION
    type = type[false]
    variableDeclaratorList = variableDeclarators
    endToken:SEMICOLON
    {
        //EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seImplementationFieldDeclaration
            = seCreator.createSE_ImplementationFieldDeclaration
                        ( getPosition( startToken )
                        , getPosition( endToken )
                        , javadoc
                        , optAnnotationList
                        , type
                        , variableDeclaratorList
                        );
    }
;

/**
 * Implementation initializer.
 */
seImplementationInitializer[ Token startToken
                           , List<Annotation> optAnnotationList
                           , List<Token> optModifierList
                           ]
    returns [SE_ImplementationInitializer seImplementationInitializer = null]
{
    Block body = null;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    IMPLEMENTATION
    body = block
    {
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seImplementationInitializer
            = seCreator.createSE_ImplementationInitializer
                        ( getPosition( startToken )
                        , body != null ? body.getEndPosition() : null
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , body
                        );
    }
;

/**
 * Mediator method declaration.
 */
seMediatorMethodDeclaration[ Token startToken
                           , List<Annotation> optAnnotationList
                           , List<Token> optModifierList
                           ]
returns [SE_MediatorMethodDeclaration seMediatorMethodDeclaration = null]
{
    TypeParameters optTypeParameters = null;
    OptionalType type = null;
    MethodHeader methodHeader = null;
    Block body = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    MEDIATOR
    optTypeParameters = optTypeParameters
    type = optionalType[false]
    methodHeader = methodHeader[optTypeParameters,type]
    {
        cTree = createTree;
        createTree( true );
    }
    (
        body = block
      | endToken:SEMICOLON
    )
    {
        createTree( cTree );
        removeInappropriates( optModifierList, methodDeclarationInappropriatesModifiers );
        // TODO: check that parameter type list is empty
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seMediatorMethodDeclaration
            = seCreator.createSE_MediatorMethodDeclaration
                        ( getPosition( startToken )
                        , body != null ? body.getEndPosition() : getPosition( endToken )
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , methodHeader
                        , body
                        );
    }
;

seUpdateBlockDeclaration[ Token startToken
                        , List<Annotation> optAnnotationList
                        , List<Token> optModifierList
                        ]
returns [SE_UpdateBlockDeclaration seUpdateBlockDeclaration = null]
{
    Block body = null;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    UPDATE
    body = block
    {
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seUpdateBlockDeclaration
            = seCreator.createSE_UpdateBlockDeclaration
                        ( getPosition( startToken )
                        , body != null ? body.getEndPosition() : null
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , body
                        );
    }
;

// --------------------------------------------------------------
//     s c e n a r i o   c l a s s
// --------------------------------------------------------------

/**
 * Scenario class declaration
 */
seScenarioClassDeclaration[ Token startToken
                          , List<Annotation> optAnnotationList
                          , List<Token> optModifierList
                          ]
    returns [SE_ScenarioClassDeclaration seScenarioClassDeclaration = null]
{
    Identifier name = null;
    TypeParameters optTypeParameters = null;
    QualifiedParameterizedIdentifier optSuperClass = null;
    List<QualifiedParameterizedIdentifier> optInterfaceList = null;
    TypeBody typeBody = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); } 
    SCENARIO CLASS
    name = identifier
    { cTree = createTree; createTree( true ); }
    optTypeParameters = optTypeParameters
    ( EXTENDS
      optSuperClass = qualifiedParameterizedIdentifier[false]
    )?
    {
        createTree( true );
    }
    (  IMPLEMENTS optInterfaceList = optInterfaceList
     | { optInterfaceList = listFactory.create()/*Type*/; }
    )
    {
        createTree( cTree );
    }
    typeBodyStartToken:LBRACE
    typeBody = typeBody[typeBodyStartToken]
    {
        Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seScenarioClassDeclaration
            = seCreator.createSE_ScenarioClassDeclaration
                        ( getPosition( startToken )
                        , getPosition( endToken )
                        , typeBody
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , name
                        , optTypeParameters
                        , optSuperClass
                        , optInterfaceList
                        );
    }
    RBRACE
;

/**
 * Scenario state block declaration
 */
seScenarioStateBlock[ Token startToken
                    , List<Annotation> optAnnotationList
                    , List<Token> optModifierList
                    ]
returns [SE_ScenarioStateBlock seScenarioStateBlock = null]
{
    Block body = null;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    STATE
    body = block
    {
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seScenarioStateBlock
            = seCreator.createSE_ScenarioStateBlock
                        ( getPosition( startToken )
                        , body != null ? body.getEndPosition() : null
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , body
                        );
    }
;

/**
 * Scenario method declaration.
 */
seScenarioMethodDeclaration[ Token startToken
                           , List<Annotation> optAnnotationList
                           , List<Token> optModifierList
                           ]
returns [SE_ScenarioMethodDeclaration seScenarioMethodDeclaration = null]
{
    Identifier name = null;
    Block body = null;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    SCENARIO
    name = identifier
    LPARENT RPARENT
    body = block
    {
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );

        seScenarioMethodDeclaration
            = seCreator.createSE_ScenarioMethodDeclaration
                        ( getPosition( startToken )
                        , body != null ? body.getEndPosition() : null
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , name
                        , body
                        );
    }
;

seReactionMediatorClassDeclaration[ Token startToken
                                  , List<Annotation> optAnnotationList
                                  , List<Token> optModifierList
                                  ]
    returns [SE_ReactionMediatorClassDeclaration seReactionMediatorClassDeclaration = null]
{
    Identifier name = null;
    TypeParameters optTypeParameters = null;
    QualifiedParameterizedIdentifier optSuperClass = null;
    List<QualifiedParameterizedIdentifier> optInterfaceList = null;
    TypeBody typeBody = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); } 
    REACTION MEDIATOR CLASS
    name = identifier
    {
        cTree = createTree; createTree( true );
    }
    optTypeParameters = optTypeParameters
    ( EXTENDS
      optSuperClass = qualifiedParameterizedIdentifier[false]
    )?
    {
        createTree( true );
    }
    (  IMPLEMENTS optInterfaceList = optInterfaceList
     | { optInterfaceList = listFactory.create()/*Type*/; }
    )
    {
        createTree( cTree );
    }
    typeBodyStartToken:LBRACE
    typeBody = typeBody[typeBodyStartToken]
    {
        Token endToken = LT(1);
        removeInappropriates( optModifierList, methodDeclarationInappropriatesModifiers );
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seReactionMediatorClassDeclaration
            = seCreator.createSE_ReactionMediatorClassDeclaration
                        ( getPosition( startToken )
                        , getPosition( endToken )
                        , typeBody
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , name
                        , optTypeParameters
                        , optSuperClass
                        , optInterfaceList
                        );
    }
    RBRACE
;

seReactionMediatorMethodDeclaration[ Token startToken
                           , List<Annotation> optAnnotationList
                           , List<Token> optModifierList
                           ]
returns [SE_ReactionMediatorMethodDeclaration seReactionMediatorMethodDeclaration = null]
{
    TypeParameters optTypeParameters = null;
    OptionalType resultType = null;
    MethodHeader methodHeader = null;
    Block body = null;
    boolean cTree = createTree;
    String javadoc = null;
}
:
    { javadoc = javadocFilter.getComment( startToken ); }
    REACTION MEDIATOR
    optTypeParameters = optTypeParameters
    resultType = optionalType[false]
    methodHeader = methodHeader[optTypeParameters,resultType]
    {
        cTree = createTree;
        createTree( true );
    }
    (
        body = block
      | endToken:SEMICOLON
    )
    {
        createTree( cTree );

        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seReactionMediatorMethodDeclaration
            = seCreator.createSE_ReactionMediatorMethodDeclaration
                        ( getPosition( startToken )
                        , body != null ? body.getEndPosition() : getPosition( endToken )
                        , javadoc
                        , optAnnotationList
                        , modifiers
                        , methodHeader
                        , body
                        );
    }
;

statement returns [Statement stmnt = null]
:
    stmnt = seIterateBase
  | stmnt = seCoverageIterate
  | stmnt = seFilterStatement
  | stmnt = seSerialize
  | stmnt = seChoice
  | ( modifiers[null,null] STATE ) => stmnt = seStateVariableDeclaration
  | stmnt = seBranchStatement
  | stmnt = seMarkStatement
  | stmnt = seIrrelevantStatement
  | stmnt = seTautology
  | stmnt = seImplementationStatement
  | stmnt = seUpdateStatement
  | stmnt = seExecuteStatement
  | ( PRE LBRACKET )=> stmnt = seCheckedPreStatement
  | stmnt = seChannelizeEventsStatement
  | ( EXPECT NEW SEMICOLON ) => stmnt = seExpectInitializeStatement
  | ( DO LBRACKET ) => stmnt = seExpectEventAfterStimulusStatement
  | stmnt = seGeneralExpectEventStatement
  | stmnt = javaStatement
;

returnStatement returns [Statement stmnt = null]
{
    Expression expr = null;
    Expression comment = null;
}
:
    // Return an expression
    rtt:RETURN ( expr = expression ( COLON comment = expression )? )? returnEndToken:SEMICOLON
    {
      if( comment != null )
      {
        stmnt = seCreator.createSE_ReturnStatement( getPosition( rtt ), getPosition( returnEndToken ), expr, comment );
      } else {
        stmnt = creator.createReturnStatement( getPosition( rtt ), getPosition( returnEndToken ), expr );
      }
    }
;

seIterateBase returns [SE_IterateBase stmnt = null]
:
    ( ITERATE LPARENT modifiers[null,null] type[false] identifier COLON ) => stmnt = seCollectionIterate
  | stmnt = seIterate
;

seIterate returns [SE_Iterate seIterate = null]
{
    Type type = null;
    VariableDeclarator declarator = null;
    Expression condition = null;
    List<Expression> optForUpdateStatementExpressionList = null;
    Expression optFilterExpression = null;
    Statement body = null;
}
:
    startToken:ITERATE
    LPARENT
    type = type[false]
    declarator = variableDeclarator
    SEMICOLON
    condition = expression
    SEMICOLON
    optForUpdateStatementExpressionList = optExpressionList
    ( SEMICOLON ( optFilterExpression = expression )? )?
    RPARENT
    body = statement
    {
        seIterate = seCreator.createSE_Iterate
                              ( getPosition( startToken )
                              , type
                              , optFilterExpression
                              , body
                              , declarator
                              , condition
                              , optForUpdateStatementExpressionList
                              );
    }
;

seCollectionIterate returns [SE_CollectionIterate seCollectionIterate = null]
{
    Type type = null;
    Identifier name = null;
    Expression collection = null;
    Expression optFilterExpression = null;
    Statement body = null;
}
:
    startToken:ITERATE
    LPARENT
    type = type[false]
    name = identifier
    COLON
    collection = expression
    ( SEMICOLON ( optFilterExpression = expression )? )?
    RPARENT
    body = statement
    {
        seCollectionIterate = seCreator.createSE_CollectionIterate
                                        ( getPosition( startToken )
                                        , type
                                        , optFilterExpression
                                        , body
                                        , name
                                        , collection
                                        );
    }
;

seCoverageSpecifier returns [SE_CoverageSpecifier coverageSpecifier = null]
{
    Expression lowerCoverageBound = null;
    Expression upperCoverageBound = null;
}
:
    lowerCoverageBound = expression
    (   COMMA
        upperCoverageBound = expression
        {
            coverageSpecifier = seCreator.createSE_IntegralTypeCoverageSpecifier
                                          ( lowerCoverageBound, upperCoverageBound );
        }
      |
        {
               coverageSpecifier = seCreator.createSE_ExpressionCoverageSpecifier
                                             ( lowerCoverageBound );
        }
    )
;

seCoverageIterate returns [SE_CoverageIterate seCoverageIterate = null]
{
    Token startToken = null;
    boolean casual = false;
    SE_CoverageSpecifier optCoverageSpecifier = null;
    SE_IterateBase iterateBase = null;
}
:
    { startToken = LT(1); }
    ( CASUAL { casual = true; } )?
    COVERAGE (  LBRACKET optCoverageSpecifier = seCoverageSpecifier RBRACKET )? COLON
    iterateBase = seIterateBase
    {
        seCoverageIterate = seCreator.createSE_CoverageIterate
                                      ( getPosition( startToken )
                                      , casual
                                      , optCoverageSpecifier
                                      , iterateBase
                                      );
    }
;

seFilterStatement returns [SE_FilterStatement seFilterStatement = null]
:
    startToken:FILTER
    (   seFilterStatement = seExpressionFilterStatement[startToken]
      | seFilterStatement = seMethodInvocationFilterStatement[startToken]
    )
;

seExpressionFilterStatement[Token startToken] returns [SE_FilterStatement seFilterStatement = null]
{
    Expression expr = null;
}
:
    LBRACKET
    expr = expression
    RBRACKET
    {
        Token endToken = LT(1);
        seFilterStatement = seCreator.createSE_ExpressionFilterStatement( getPosition( startToken ), getPosition( endToken ), expr );
    }
    SEMICOLON
;

methodInvocationExpression returns [MethodInvocation mi = null]
{
    Expression expr = null;
}
:
    expr = expression // MethodInvocation
    {
        if( createTree )
        {
            if( expr instanceof MethodInvocation )
            {
                mi = (MethodInvocation)expr;
            } else {
                mbox.error( file
                          , startVisitor.getStartPosition( expr )
                          , new Messages.MethodInvocationExpressionMessage()
                          );
                stopCreateTree();
            }
        }
    }
;

seMethodInvocationFilterStatement[Token startToken] returns [SE_FilterStatement seFilterStatement = null]
{
    SE_InvocationResultStorage optResultStorage = null;
    MethodInvocation mi = null;
}
:
    COLON
    ( ( seInvocationResultStorage )=>
      optResultStorage = seInvocationResultStorage )?
    mi = methodInvocationExpression
    {
        Token endToken = LT(1);
        seFilterStatement = seCreator.createSE_MethodInvocationFilterStatement
                                      ( getPosition( startToken ), getPosition( endToken ), optResultStorage, mi );
    }
    SEMICOLON
;


seInvocationResultStorage
    returns [SE_InvocationResultStorage seInvocationResultStorage = null]
:
      ( seAssignmentResultStorage )=>
      seInvocationResultStorage = seAssignmentResultStorage
    | seInvocationResultStorage = seVariableDeclarationResultStorage
;

seVariableDeclarationResultStorage
    returns [SE_VariableDeclarationResultStorage seVariableDeclarationResultStorage = null]
{
    Token startToken = null;
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    Type type = null;
    VariableDeclaratorId name = null;
}
:
    {
        optAnnotationList = listFactory.create()/*Annotation*/;
        optModifierList = listFactory.create()/*Modifier*/;
        startToken = LT(1);
    }
    modifiers[optAnnotationList,optModifierList]
    type = type[false]
    name = variableDeclaratorId
    {
        Token endToken = LT(1);
        EnumSet<Modifier> modifiers = TokenUtils.getModifiers( optModifierList );
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );
        seVariableDeclarationResultStorage
            = seCreator.createSE_VariableDeclarationResultStorage
                        ( getPosition( startToken )
                        , getPosition( endToken )
                        , optAnnotationList
                        , modifiers
                        , type
                        , name
                        );
    }
    ASSIGN
;

seAssignmentResultStorage returns [SE_AssignmentResultStorage seAssignmentResultStorage = null]
{
    Token endToken = null;
    QualifiedIdentifier qid = null;
    AssignmentOperator operator = null;
}
:
    qid = qualifiedIdentifier
    { endToken = LT(1); }
    operator = assignmentOperator
    {
        seAssignmentResultStorage = seCreator.createSE_AssignmentResultStorage( getPosition( endToken ), creator.createExpressionName( qid ), operator );
    }
;

seSerialize returns [SE_Serialize seSerialize = null]
{
    Block block = null;
    List<Block> optBlockList = null;
    boolean cTree = createTree;
}
:
{
    createTree( true );
    optBlockList = listFactory.create();
    createTree( cTree );
}
    startToken:SERIALIZE
    LBRACE
    ( { createTree( true ); }
        block = block
        {
            listFactory.add( optBlockList, block );
        }
     )*
    {
        Token endToken = LT(1);
        createTree( cTree );
        seSerialize = seCreator.createSE_Serialize( getPosition( startToken ), getPosition( endToken ), optBlockList );
    }
    RBRACE
;

seChoice returns [SE_Choice seChoice = null]
{
    Block block = null;
    List<Block> optBlockList = null;
    boolean cTree = createTree;
}
:
{
    createTree( true );
    optBlockList = listFactory.create();
    createTree( cTree );
}
    startToken:CHOICE
    LBRACE
    ( { createTree( true ); }
        block = block
        {
            listFactory.add( optBlockList, block );
        }
    )*
    {
        Token endToken = LT(1);
        createTree( cTree );
        seChoice = seCreator.createSE_Choice( getPosition( startToken ), getPosition( endToken ), optBlockList );
    }
    RBRACE
;

seStateVariableDeclaration returns [SE_StateVariableDeclaration seStateVariableDeclaration = null]
{
    List<Annotation> optAnnotationList = null;
    List<Token> optModifierList = null;
    Type type = null;
    List<VariableDeclarator> variableDeclaratorList = null;
    Token startToken = null;
}
:
    {
        optAnnotationList = listFactory.create();
        optModifierList = listFactory.create();
        startToken = LT(1);
    }
    modifiers[optAnnotationList,optModifierList]
    STATE
    type = type[false]
    variableDeclaratorList = variableDeclarators
    {
        Token endToken = LT(1);
        TokenUtils.checkAllowableModifiers( optModifierList, mbox, file );

        seStateVariableDeclaration
            = seCreator.createSE_StateVariableDeclaration
                        ( getPosition( startToken )
                        , getPosition( endToken )
                        , optAnnotationList
                        , type
                        , variableDeclaratorList
                        );
    }
    SEMICOLON
;

seBranchStatement returns [SE_BranchStatement seBranchStatement = null]
{
    Identifier name = null;
    Expression optDescription = null;
    SE_UpdateStatement optUpdate = null;
}
:
    startToken:BRANCH
    name = identifier
    ( LPARENT optDescription = expression RPARENT )?
    (
       SEMICOLON
     | optUpdate = seUpdateStatement
    )
    {
        Token endToken = LT(1);
        seBranchStatement = seCreator.createSE_BranchStatement
                                      ( getPosition( startToken )
                                      , getPosition( endToken )
                                      , name
                                      , optDescription
                                      , optUpdate
                                      );
    }
;


seMarkStatement returns [SE_MarkStatement seMarkStatement = null]
{
    Identifier name = null;
}
:
    startToken:MARK
    ( name = identifier )?
    s:STRING_LITERAL
    {
        Token endToken = LT(1);
        seMarkStatement = seCreator.createSE_MarkStatement
                                    ( getPosition( startToken )
                                    , getPosition( endToken )
                                    , name
                                    , s.getText()
                                    );
    }
    SEMICOLON
;

seIrrelevantStatement returns [SE_IrrelevantStatement seIrrelevantStatement = null]
{
    Statement statement = null;
}
:
    startToken:IRRELEVANT
    statement = statement
    {
        seIrrelevantStatement = seCreator.createSE_IrrelevantStatement( getPosition( startToken ), statement );
    }
;

seTautology returns [SE_TautologyStatement seTautologyStatement = null]
{
    Expression expression = null;
}
:
    startToken:TAUTOLOGY
    expression = expression
    {
        Token endToken = LT(1);
        seTautologyStatement = seCreator.createSE_TautologyStatement( getPosition( startToken ), getPosition( endToken ), expression );
    }
    SEMICOLON
;

seImplementationStatement returns [SE_ImplementationStatement seImplementationStatement = null]
{
    QualifiedIdentifier exceptionType = null;
    List<QualifiedIdentifier> optExceptionList = null;
    Statement statement = null;
    boolean cTree = createTree;
}
:
    startToken:IMPLEMENTATION
    {
        createTree( true );
        optExceptionList = listFactory.create()/*QualifiedParameterizedIdentifier*/;
    }
    ( THROWS
      exceptionType = qualifiedIdentifier
      {
        listFactory.add( optExceptionList, exceptionType );
        createTree( true );
      }
      ( COMMA
        exceptionType = qualifiedIdentifier
        {
            listFactory.add( optExceptionList, exceptionType );
            createTree( true );
        }
      )*
    )?
    {
        createTree( cTree );
    }
    statement = statement
    {
        seImplementationStatement = seCreator.createSE_ImplementationStatement( getPosition( startToken )
                                                                              , statement
                                                                              , optExceptionList
                                                                              );
    }
;

seUpdateStatement returns [SE_UpdateStatement seUpdateStatement = null]
{
    Statement statement = null;
    SE_InvocationResultStorage optResultStorage = null;
    MethodInvocation optReactionInvocation = null;
}
:
    startToken:UPDATE
    (
        LBRACKET
        ( ( seInvocationResultStorage )=>
        optResultStorage = seInvocationResultStorage )?
        optReactionInvocation = methodInvocationExpression
        RBRACKET
    )?
    statement = statement
    {
        seUpdateStatement = seCreator.createSE_UpdateStatement( getPosition( startToken )
                                                              , optResultStorage
                                                              , optReactionInvocation
                                                              , statement
                                                              );
    }
;

seExecuteStatement returns [SE_ExecuteStatement seExecuteStatement = null]
{
    Statement statement = null;
}
:
    startToken:EXECUTE
    statement = statement
    {
        seExecuteStatement = seCreator.createSE_ExecuteStatement( getPosition( startToken ), statement );
    }
;

seCheckedPreStatement returns [SE_CheckedPreStatement seCheckedPreStatement = null]
{
    SE_InvocationResultStorage optResultStorage = null;
    MethodInvocation mi = null;
    Statement optStatement = null;
    boolean cTree = createTree;
}
:
    startToken:PRE
    LBRACKET
    ( ( seInvocationResultStorage )=>
      optResultStorage = seInvocationResultStorage )?
    mi = methodInvocationExpression
    RBRACKET
    {
        cTree = createTree;
        createTree( true );
    }
    (   COLON optStatement = statement
      | endToken:SEMICOLON
    )
    {
        createTree( cTree );

        Position endPosition =   endToken == null && optStatement != null
                               ?   endVisitor.getEndPosition( optStatement )
                                 : getPosition( endToken );
        seCheckedPreStatement = seCreator.createSE_CheckedPreStatement
                                          ( getPosition( startToken )
                                          , endPosition
                                          , optResultStorage
                                          , mi
                                          , optStatement
                                          );
    }
;

seChannelizeEventsStatement returns [SE_ChannelizeEventsStatement seChannelizeEventsStatement = null]
{
    Java5.Expression optChannelID = null;
    boolean isDefaultChannel = false;
    Java5.Block block = null;
}
:
    startToken:LTHN c1:COLON
    {
        if(    c1.getLine()  != startToken.getLine()
            || c1.getColumn() - startToken.getColumn() != 1
          )
        {
            mbox.error( file
                      , getPosition( startToken )
                      , new Messages.IncorrectChannelIdDelimeter()
                      );
        }
    }
    (
        optChannelID = expression
      | (
          DEFAULT
          { isDefaultChannel = true; }
        )
    )
    c2:COLON endToken:GTHN
    {   if(    endToken.getLine()  != c2.getLine()
            || endToken.getColumn() - c2.getColumn() != 1
          )
        {
            mbox.error( file
                      , getPosition( c2 )
                      , new Messages.IncorrectChannelIdDelimeter()
                      );
        }
    }
    block = block
    {
        seChannelizeEventsStatement
            = seCreator.createSE_ChannelizeEventsStatement( getPosition( startToken ), optChannelID, isDefaultChannel, block );
    }
;

seExpectInitializeStatement returns [SE_ExpectInitializeStatement seExpectInitializeStatement = null]
{
}
:
    startToken:EXPECT NEW endToken:SEMICOLON
    {
        seExpectInitializeStatement = seCreator.createSE_ExpectInitializeStatement( getPosition( startToken )
                                                                                  , getPosition( endToken ) 
                                                                                  );
    }
; 

seExpectEventAfterStimulusStatement returns [SE_ExpectEventAfterStimulusStatement seExpectEventAfterStimulusStatement = null]
{
    MethodInvocation mi = null; 
    Expression e = null;
    SE_EventExpectationElement ev = null;
    List<SE_EventExpectationElement> evList = null;
}
:
    startToken:DO
    LBRACKET mi=methodInvocationExpression RBRACKET
    EXPECT
    ( FOR LBRACKET e=expression RBRACKET )?
    LBRACE
    { evList = listFactory.create()/*SE_EventExpectationElement*/; }
    ( ev = seEventExpectationElement { listFactory.add( evList, ev ); } )+
    endToken:RBRACE
    {
        seExpectEventAfterStimulusStatement 
            = seCreator.createSE_ExpectEventAfterStimulusStatement(
                          getPosition( startToken )
                        , getPosition( endToken )
                        , e
                        , evList 
                        , mi
                        );
    }
;

/**
 * SE_GeneralExpectEventStatement ::= "expect"
 *                                    ( "new" )?
 *                                    ( "for" "[" Expression "]" )?
 *                                    "{"
 *                                    ( SE_EventExpectationElement )+
 *                                    "}"
 *                                ;
 */
seGeneralExpectEventStatement returns [SE_GeneralExpectEventStatement seGeneralExpectEventStatement = null]
{ 
    boolean expectingNew = false;
    Expression e = null;
    SE_EventExpectationElement ev = null;
    List<SE_EventExpectationElement> evList = null;
}
:
    startToken:EXPECT
    ( NEW { expectingNew = true; } )?
    ( FOR LBRACKET e=expression RBRACKET )?
    LBRACE
    { evList = listFactory.create()/*SE_EventExpectationElement*/; }
    ( ev = seEventExpectationElement { listFactory.add( evList, ev ); } )+
    endToken:RBRACE
    {
        seGeneralExpectEventStatement 
            = seCreator.createSE_GeneralExpectEventStatement(
                          getPosition( startToken )
                        , getPosition( endToken )
                        , e
                        , evList 
                        , expectingNew
                        );
    }
;

seEventExpectationElement returns [SE_EventExpectationElement seEventExpectationElement = null]
{
    SE_StopExpectSelector e = null;
    List<SE_StopExpectSelector> el = null;
    BlockStatement s = null;
    List<BlockStatement> sl = null;
}
:
    { el = listFactory.create(); }
    ( e = seStopExpectSelector { listFactory.add( el, e ); } )+
    { sl = listFactory.create(); }
    ( s = blockStatement { listFactory.add( sl, s ); } )+
    {
        seEventExpectationElement = seCreator.createSE_EventExpectationElement( el, sl );
    }
;
 
seStopExpectSelector returns [SE_StopExpectSelector seStopExpectSelector = null]
{
    Expression e = null;
}
:
    s1:CASE 
    e = expression e2:COLON 
    { seStopExpectSelector = seCreator.createSE_StopExpectOnPattern( getPosition(s1), getPosition(e2), e ); }
  | s3:TIMEOUT e3:COLON 
    { seStopExpectSelector = seCreator.createSE_StopExpectOnTimeout( getPosition(s3), getPosition(e3) ); }
;

primary returns [Primary primary_expr = null]
:
    primary_expr = seMediatorCreationExpression
  | primary_expr = seEventPatternExpression
  | primary_expr = javaPrimary
;

seMediatorCreationExpression
returns [SE_MediatorCreationExpression seMediatorCreationExpression = null]
{
    Type type = null;
    SE_MediatorCreationParameter param = null;
    List<SE_MediatorCreationParameter> optParameterList = null;
    boolean cTree = createTree;
}
:
{
    createTree( true );
    optParameterList = listFactory.create();
    createTree( cTree );
}
    startToken:MEDIATOR
    type = type[false]
    LPARENT
    ( { cTree = createTree; createTree( true ); }
        param = seMediatorCreationParameter
        {
            listFactory.add( optParameterList, param );
            createTree( cTree );
        }
     )?
    (
      { createTree( true ); }
      COMMA
      param = seMediatorCreationParameter
      {
        listFactory.add( optParameterList, param );
      }
    )*
    {
        Token endToken = LT(1);
        createTree( cTree );
        seMediatorCreationExpression = seCreator.createSE_MediatorCreationExpression
                                       ( getPosition( startToken )
                                       , getPosition( endToken )
                                       , type
                                       , optParameterList
                                       );
    }
    RPARENT
;

seMediatorCreationParameter
returns [SE_MediatorCreationParameter seMediatorCreationParameter = null]
{
    Identifier name = null;
    Expression value = null;
}
:
    name = identifier
    ASSIGN
    value = expression
    {
        seMediatorCreationParameter = seCreator.createSE_MediatorCreationParameter
                                                ( name, value );
    }
;

unaryExpressionNotPlusMinus returns [UnaryExpressionNotPlusMinus expr = null]
:
    expr = sePreExpression
  | expr = seThrownExpression
  | expr = javaUnaryExpressionNotPlusMinus
;

sePreExpression returns [SE_PreExpression sePreExpression = null]
{
    UnaryExpression expr = null;
}
:
    startToken:PRE
    expr = unaryExpression
    {
        sePreExpression = seCreator.createSE_PreExpression( getPosition( startToken ), expr );
    }
;

seThrownExpression returns [SE_ThrownExpression seThrownExpression = null]
{
}
: t:THROWN { seThrownExpression = seCreator.createSE_ThrownExpression( getPosition( t ) ); }
;

/**
 * TDL:
 *    SE_EventPatternExpression
 *    |- SE_PrimaryEventPatternExpression
 *    '- SE_QualifiedEventPatternExpression
 */
seEventPatternExpression returns [SE_EventPatternExpression seEventPatternExpression = null]
{
    Primary pex = null;
    TypeArguments ta = null;
    QualifiedIdentifier qid = null;
    Identifier name = null;
    Expression e = null;
    SE_PatternArgument p = null;
    List<SE_PatternArgument> l = null;
}
:
    EVENT
    (
      (
          ( qualifiedIdentifier SHARP ) =>
          qid = qualifiedIdentifier
        | pex = primary
      )
    )?
    SHARP 
    ta = optTypeArguments[false,false]
    name = identifier
    ( LTHN COLON e = expression COLON GTHN )?
    LPARENT
    { l = listFactory.create(); }
    (
              p = sePatternArgument { listFactory.add( l, p ); }
      ( COMMA p = sePatternArgument { listFactory.add( l, p ); } )*
    )?
    endToken:RPARENT
    {
        if( pex != null )
        {
            seEventPatternExpression 
                = seCreator.createSE_PrimaryEventPatternExpression
                            ( getPosition(endToken)
                            , ta
                            , name
                            , e
                            , l
                            , pex 
                            );
        } else {
            seEventPatternExpression 
                = seCreator.createSE_QualifiedEventPatternExpression
                            ( getPosition(endToken)
                            , ta
                            , name
                            , e
                            , l
                            , qid 
                            );
        }
    }
;

sePatternArgument returns [SE_PatternArgument sePatternArgument = null]
{
    QualifiedIdentifier qid = null;
    Type t = null;
    int dim = 0;
    boolean va = false;
    Expression e = null;
    Identifier id = null;
}
:
      ( qualifiedIdentifier (VARARG)? (variableDeclaratorId)? ) =>
      qid = qualifiedIdentifier ( VARARG { va = true; } )? ( id = identifier ( LBRACKET  endToken1:RBRACKET { dim++; } )* )?
      {
        if( va || id != null )
        {
            List<ParameterizedIdentifier> qpidList = listFactory.create();
            for( Identifier qid_i : qid.getIdentifierList() )
            {
                listFactory.add( qpidList, creator.createParameterizedIdentifier
                                                   ( creator.createIdentifier( qid_i.getPosition()
                                                                             , qid_i.getValue()
                                                                             )
                                                   , null 
                                                   ) 
                               );
            }
            t = creator.createQualifiedParameterizedIdentifier( qpidList );
            if( dim > 0 ) t = creator.createArrayType( getPosition( endToken1 ), t, dim );
            sePatternArgument = seCreator.createSE_TypedPatternArgument( t, va, id );
        } else {
            sePatternArgument = seCreator.createSE_QIDPatternArgument( qid );
        }
      }   
    | ( primitiveType ( LBRACKET  RBRACKET )* DOT CLASS )=>
      e = expression
      { sePatternArgument = seCreator.createSE_ValuePatternArgument( e ); }
    | t = type[false] ( VARARG { va = true; } )? ( id = identifier ( LBRACKET  endToken2:RBRACKET { dim++; } )* )?
      { 
          if( dim > 0 ) t = creator.createArrayType( getPosition( endToken2 ), t, dim );
          sePatternArgument = seCreator.createSE_TypedPatternArgument( t, va, id ); 
      }      
    | e = expression 
      { sePatternArgument = seCreator.createSE_ValuePatternArgument( e ); }
;
